package Net::Kismet;

require v5.6.0;
require Exporter;
use strict;
use warnings;
use Socket qw(:DEFAULT :crlf);
use Carp;

our @ISA = qw(Exporter);
our @EXPORT = qw(KismetRun);
our @EXPORT_OK = qw(KismetInit KismetLoop);
our %EXPORT_TAGS = (manual => [qw(KismetInit KismitLoop)], 
		    callback => [qw(KismetRun)]);
our $VERSION = 0.04;

our $TIMEOUT = 0.5;
our $ACK_TIMEOUT = 5;

sub new {
	my ($class, $host, $port) = @_;
	$host ||= 'localhost';
	$port ||= 2501;
	my $self = {'_host', $host, '_port', $port};
	my $ip = inet_aton($host)
		or croak("Cannot resolve host name: $!\n");
	$self->{'_addr'} = sockaddr_in($port, $ip);
	socket($self->{'_socket'}, PF_INET, SOCK_STREAM, getprotobyname('tcp'))
		or croak("Cannot get a socket: $!\n");
	select((select($self->{'_socket'}), $| = 1)[0]);
	$self->{'_startup'} = 1;
	$self->{'_cid'} = 1;
	$self->{'_block'} = (stat($self->{'_socket'}))[11] || 4096;
	$self->{'_buffer'} = "";
	return bless $self, $class;
}

sub register {
	my ($self, $proto, $callback, $opts) = @_;
	$opts ||= '*';
	if (!defined($proto)) {
		croak("No protocall given to register callback");
	}
	if (ref($callback) ne 'CODE') {
		croak("Attempted to register no/invalid callback for $proto\n");
	}
	if ($self->{'_startup'}) {
		$self->preregister($proto, $callback, $opts);
		return;
	}
	$proto = uc $proto;
	if (exists $self->{'_capmap'}->{$proto}) {
		my @o = ();
		if ($opts eq '*') {
			@o = keys %{$self->{'_capmap'}->{$proto}};
		} else {
			$opts = lc $opts;
			foreach (split /,/, $opts) {
				if (exists $self->{'_capmap'}->{$proto}->{$_}) {
					push @o, $_;
				}
			}
		}
		$self->{'_callback'}->{$proto} = [$callback, \@o];
		$self->command('ENABLE', "$proto *");
	} else {
		carp("$proto not a valid protocall for this server\n");
	}
}

sub timer_register {
	my ($self, $callback, $sec) = @_;
	$sec ||= 1;
	if (ref($callback) ne 'CODE') {
		croak("Attempted to register no/invalid timer callback for\n");
	}
	$self->{'_timer'} = [$callback, $sec, time + $sec];
}

sub KismetRun {
	KismetInit(@_);
	while(defined KismetLoop(@_)) {}
}

sub KismetInit {
	$_->prestartup() foreach @_;
}

sub KismetLoop {
	my $rin = '';
	my $alive = 0;
	foreach (@_) {
		$_->timer();
		if (!$_->{'_shutdown'}) {
			vec ($rin, fileno($_->{'_socket'}) , 1) = 1;
			$alive++;
			$_->checkack();
		}
	}
	return unless $alive;

	my $ready = 0;
	if ($ready = select(my $rout = $rin, undef, undef, $TIMEOUT)) {
		foreach (@_) {
			if (vec $rout, fileno($_->{'_socket'}), 1) {
				$_->readsocket();
			}
		}
	}
	return $ready;
}

sub DESTROY {
	my $self = shift;
	shutdown $self->{'_socket'}, 2;
}

sub prestartup {
	my $self = shift;
	connect($self->{'_socket'}, $self->{'_addr'})
		or croak("Cannot connect to kismet server $self->{'_host'}:$self->{'_port'}: $!\n");
}

sub poststartup {
	my $self = shift;
	my $rin = my $rout = '';
	vec ($rin, fileno($self->{'_socket'}), 1) = 1;
	while (my $pr = shift @{$self->{'_precallback'}}) {
		$self->register(@$pr);
	}
	while (my $bl = shift @{$self->{'_backlog'}}) {
		$self->dispatch(@$bl);
	}
}

sub readsocket {
	my $self = shift;
	my $rin = '';
	vec ($rin, fileno($self->{'_socket'}), 1) = 1;
       {
	my $read = sysread $self->{'_socket'}, $self->{'_buffer'}, $self->{'_block'}, length($self->{'_buffer'});
	if (!defined $read) {
		$self->{'_shutdown'} = 1;
		carp("Read error: $!\n");
		return;
	} elsif ($read == 0) {
		$self->{'_shutdown'} = 1;
		return;
	}

	my ($p, $a);
	while(($p, $a) = parse(\$self->{'_buffer'}), defined $p) {
		if ($p eq 'ACK') {
			$self->clearack($a->[0]);
		}
		if ($self->{'_startup'}) {
			if ($p eq 'PROTOCOLS') {
				$self->{'_startup'} = 0;
				foreach (split /,/, $a->[0]) {
					$self->command('CAPABILITY', $_);
					$self->{'_startup'}++;
				}
			}
			elsif ($p eq 'CAPABILITY') {
				my $x = 0;
				my $uline = "";
				foreach (split /,/, $a->[1]) {
					$self->{'_capmap'}->{$a->[0]}->{$_} = $x++;
					$uline = $_;
				}
				if ($x == 1) {
					$self->{'_capmap'}->{$a->[0]}->{$uline} = '*';
				}
				$self->{'_startup'}--;
			}
			push @{$self->{'_backlog'}}, [$p, $a];
			unless ($self->{'_startup'}) {
				$self->poststartup();
			}
		} else {
			$self->dispatch($p, $a);
		}
	}
	redo if select(my $rout = $rin, undef, undef, 0);
       }
}

sub preregister {
	my ($self, $proto, $callback, $opts) = @_;
	push @{$self->{'_precallback'}}, [$proto, $callback, $opts];
}

sub timer {
	my $self = shift;
	return unless defined $self->{'_timer'};
	if ($self->{'_timer'}->[2] <= time) {
		$self->{'_timer'}->[0]($self);
		$self->{'_timer'}->[2] = time + $self->{'_timer'}->[1];
	}
}

sub dispatch {
	my ($self, $proto, $args) = @_;
	if (exists $self->{'_callback'}->{$proto}) {
		my %hash = ();
		foreach (@{$self->{'_callback'}->{$proto}->[1]}) {
			my $id = $self->{'_capmap'}->{$proto}->{$_};
			if ($id eq '*') {
				$hash{$_} = join " ", @$args;
			} else {
				$hash{$_} = $args->[$id];
			}
		}
		$self->{'_callback'}->{$proto}->[0](\%hash, $self);
	}
}

sub writ {
	my ($self, $string) = @_;
	my $writ = 0;
	while ($writ < length($string)) {
		my $wrote = syswrite $self->{'_socket'}, $string, $self->{'_block'}, $writ;
		unless (defined $wrote) {
			$self->{'_shutdown'} = 1;
			carp("Write failed: $!\n");
			return;
		}
		$writ += $wrote;
	}
}

sub command {
	my ($self, $command, $args) = @_;
	push @{$self->{'_acks'}}, [$self->{'_cid'}, $command, $args, time];
	$self->writ('!' . $self->{'_cid'}++ . " $command $args$CRLF");
}

sub clearack {
	my ($self, $ack) = @_;
	my $i;
	for ($i = 0; $i <= $#{$self->{'_acks'}}; $i++) {
		last if $self->{'_acks'}->[$i]->[0] == $ack;
	}
	splice @{$self->{'_acks'}}, $i, 1;
}

sub checkack {
	my $self = shift;
	my $cutoff = time - $ACK_TIMEOUT;
	for (my $i = 0; $i <= $#{$self->{'_acks'}}; $i++) {
		if ( $self->{'_acks'}->[$i]->[3] <= $cutoff )  {
			writ("!$self->{'_acks'}->[$i]->[0] $self->{'_acks'}->[$i]->[1] $self->{'_acks'}->[$i]->[2]$CRLF");
			$self->{'_acks'}->[$i]->[3] = time;
		}
	}
}

sub parse {
	my $raw = shift;
	my $proto = '';
	my @args = ();

###NOT FINISHED
	if ($$raw !~ /^\*/) {
		my $ind = index $$raw, "\n*";
		if ($ind == -1) {
			$$raw = '';
			return undef;
		}
		substr $$raw, 0, $ind+1, '';
	}
#*goobiei\001 \001dafe\n*rebok\001\n*TEST: bla\n
###END

	my $pos = my $end = 0;
	my $string = '';

	{
	$end = index $$raw, "\n", $pos;
	if ($end == -1) {
		return;
	}
	$string = substr $$raw, 0, $end+1;
	my $c = $string =~ tr/\001//;
	if ($c % 2) {
		$pos = $end+1;
		redo;
	}
	}

	substr $$raw, 0, $end+1, '';

	if ($string =~ s/^\*([A-Z]*): //) {
		chomp $string;
		$proto = $1;
		my @x = split / /, $string;
		for (my $i=0; $i <= $#x; $i++) {
			if ($x[$i] =~ /^\001/) {
				my $bits = $x[$i];
				$bits .= " " . $x[++$i]	while length($bits) == 1 || $bits !~ /\001$/;
				push @args, substr($bits, 1, -1);
			} else {
				push @args, $x[$i];
			}
		}
		return $proto, \@args;		
	}
	return;	
}

1;

__END__

=head1 NAME

Net::Kismet - Client module for Kismet

=head1 SYNOPSIS

 use Net::Kismet;
 $k = new Net::Kismet('localhost', 2501);
 $k->register('info', \&infohandler, '*');
 $k->register('network', \&networkhandler, 'ssid,wep,channel');
 $k->timer_register(\&timer, 5);
 KismetRun($k);

 use Net::Kismet qw(:manual);
 $k = new Net::Kismet('localhost', 2501);
 KismetInit($k);
 while (1) { KismetRun($k) }
  
=head1 DESCRIPTION

This module acts as a generic client for Kismet servers.

=head2 METHODS

 register()
  1st arg:
   Kismet Protocol name.

  2nd arg:
   Callback called when Protocol shows up from server.
   The callback is called with two arguments, first is a
   refrence to hash of fields, second is the object that 
   registered this callback.

  3rd arg: optional = *
   Fields passed to callback inside hash refrence. A * indicates
   request for all fields.

 timer_register()
  1st arg:
   Callback called every n seconds.
   The callback is called with one argument, the object that
   registered the callback.

  2nd arg: optional = 1
   How often callback is called (aprox)

=head2 FUNCTIONS

 KismetRun()
  Runs Net::Kismet objects, called with a list of objects.
  Returns when all servers terminate.

 KismetInit()
  Kickstarts a list of Net::Kismet objects.

 KismetLoop()
  Does one loop on all Net::Kismet objects.
  Returns number of servers that have data ready to read, undef
  when all servers terminate.

=head2 VARS

 $Net::Kismet::TIMEOUT
  How long the select loop will block.

 $Net::Kismet::ACK_TIMEOUT
  How many seconds to wait until unacknolaged commands are
  resent to server

=head2 EXPORT

 :DEFAULT
  KismetRun()

 :manual
  KismetInit()
  KismetLoop()

=head1 AUTHOR

Jason Faust, E<lt>jason.faust@hardwaremassagecircus.comE<gt>

=head1 SEE ALSO

L<perl(1)>

L<kismet(1)>

http://www.kismetwireless.net

docs/DEVEL.client from the Kismet package.

=cut
